from typing import Type

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.apps import apps
from django.db.models.base import ModelBase
from django.db.models.fields.files import ImageFieldFile
from django.utils.translation import gettext_lazy as _
from PIL import Image
from image_assets import defaults


class AssetTypeManager(models.Manager):
    def get_for_model(self, model):
        if model is None:
            return self.all()
        ct = ContentType.objects.get_for_model(model)
        return self.filter(
            models.Q(required_for=ct) | models.Q(allowed_for=ct)).distinct()

    def get_required(self, instance):
        if instance is None:
            return self.all()
        ct = ContentType.objects.get_for_model(instance)
        required = self.filter(required_for=ct)
        if instance.pk is None or isinstance(instance, ModelBase):
            return required
        existing = get_asset_model().objects.filter(
            active=True, content_type=ct, object_id=instance.pk).values(
            'asset_type')
        return required.exclude(pk__in=existing)


class AssetType(models.Model):
    JPEG = 'jpeg'
    PNG = 'png'
    FORMAT_CHOICES = (
        (JPEG, 'JPEG'),
        (PNG, 'PNG')
    )

    slug = models.SlugField(unique=True)
    format = models.CharField(
        verbose_name=_('Image Format'), max_length=4, choices=FORMAT_CHOICES)
    min_width = models.IntegerField(
        verbose_name=_('Min Width'), default=0)
    min_height = models.IntegerField(
        verbose_name=_('Min Height'), default=0)
    aspect = models.FloatField(
        verbose_name=_('Aspect'), default=0)
    accuracy = models.FloatField(
        verbose_name=_('Aspect accuracy'), default=0.01)
    max_size = models.IntegerField(
        verbose_name=_('Max file size'), default=0)

    required_for = models.ManyToManyField(
        ContentType, blank=True, related_name='required_asset_types',
        related_query_name='required_asset_types')
    allowed_for = models.ManyToManyField(
        ContentType, blank=True, related_name='allowed_asset_types',
        related_query_name='allowed_asset_types')

    class Meta:
        abstract = defaults.ASSET_TYPE_MODEL != 'image_assets.AssetType'

    objects = AssetTypeManager()

    def __str__(self):
        return self.slug

    @classmethod
    def validate_asset(cls, value: ImageFieldFile):
        asset = value.instance
        if asset.asset_type_id is None:
            # asset type not filled, no data to validate
            return
        errors = []
        asset_type: AssetType = asset.asset_type

        # validate file size
        if (asset_type.max_size and value.size and
                asset_type.max_size < value.size):
            msg = _('Image size must be not greater than %s')
            errors.append(msg % asset_type.max_size)

        # open image and validate it's content
        with Image.open(value.file) as image:  # type: Image.Image
            # internal image format
            if image.format.lower() != asset_type.format:
                msg = _('Image format must be %s')
                errors.append(msg % asset_type.format)
            # image width
            if image.width and asset_type.min_width > image.width:
                msg = _('Image width must be not less than %s')
                errors.append(msg % asset_type.min_width)
            # image height
            if image.height and asset_type.min_height > image.height:
                msg = _('Image height must be not less than %s')
                errors.append(msg % asset_type.min_height)
            if image.width and image.height and asset_type.aspect:
                image_aspect = image.width / image.height
                delta = image_aspect - asset_type.aspect
                if asset_type.accuracy == 0:
                    if image_aspect != asset_type.aspect:
                        msg = _('Image aspect must be %s')
                        errors.append(msg % asset_type.aspect)
                elif round(delta / asset_type.accuracy) > 1:
                    # round at scale of accuracy
                    msg = _('Image aspect must be %s ± %s')
                    args = (asset_type.aspect, asset_type.accuracy)
                    errors.append(msg % args)

        if errors:
            raise ValidationError(errors)


def get_asset_type_model() -> Type[AssetType]:
    app_label, model_name = defaults.ASSET_TYPE_MODEL.split('.')
    return apps.get_registered_model(app_label, model_name)


class Asset(models.Model):
    class Meta:
        abstract = defaults.ASSET_MODEL != 'image_assets.Asset'

    image = models.ImageField(validators=[AssetType.validate_asset])
    asset_type = models.ForeignKey(AssetType, models.CASCADE)
    active = models.BooleanField(default=True)

    content_type = models.ForeignKey(ContentType, models.CASCADE)
    object_id = models.IntegerField()
    related = GenericForeignKey()


def get_asset_model() -> Type[Asset]:
    app_label, model_name = defaults.ASSET_MODEL.split('.')
    return apps.get_registered_model(app_label, model_name)


class DeletedAsset(models.Model):
    image = models.ImageField()
    asset_type = models.ForeignKey(defaults.ASSET_TYPE_MODEL, models.CASCADE)

    content_type = models.ForeignKey(ContentType, models.CASCADE)
    object_id = models.IntegerField()
    related = GenericForeignKey()

    class Meta:
        abstract = defaults.DELETED_ASSET_MODEL != 'image_assets.DeletedAsset'
        unique_together = ('content_type', 'object_id')


def get_deleted_asset_model() -> Type[DeletedAsset]:
    app_label, model_name = defaults.DELETED_ASSET_MODEL.split('.')
    return apps.get_registered_model(app_label, model_name)



from django.conf import settings


__all__ = [
    'ASSET_TYPE_MODEL',
    'ASSET_MODEL',
    'DELETED_ASSET_MODEL',
]

_default_settings = {
    'ASSET_TYPE_MODEL': 'image_assets.AssetType',
    'ASSET_MODEL': 'image_assets.Asset',
    'DELETED_ASSET_MODEL': 'image_assets.DeletedAsset'
}


assets_config = dict(_default_settings)
local_config = getattr(settings, 'IMAGE_ASSETS_CONFIG', {})
for k, v in local_config.items():
    if k not in _default_settings:
        raise KeyError(k)
    assets_config[k] = v


ASSET_TYPE_MODEL = assets_config['ASSET_TYPE_MODEL']
ASSET_MODEL = assets_config['ASSET_MODEL']
DELETED_ASSET_MODEL = assets_config['DELETED_ASSET_MODEL']


from typing import Type

from django.db.models.signals import pre_delete
from django.dispatch import receiver

from image_assets import models


# noinspection PyUnusedLocal
@receiver(pre_delete, sender=models.get_asset_model())
def move_asset_file_to_deleted_asset(
        sender, *, instance: models.Asset, **kwargs):
    """ When asset is deleted, it's file is moved to deleted asset instance."""
    models.get_deleted_asset_model().objects.create(
        image=instance.image,
        content_type_id=instance.content_type_id,
        object_id=instance.object_id,
        asset_type_id=instance.asset_type_id)


# noinspection PyUnusedLocal
@receiver(pre_delete, sender=models.get_deleted_asset_model())
def delete_asset_file_for_deleted_asset(
        sender, *, instance: models.DeletedAsset, **kwargs):
    """
    When deleted asset is deleted from db, it's file is purged from storage.
    """
    instance.image.delete()



# Generated by Django 3.0.4 on 2020-03-19 15:14

from django.db import migrations, models
import django.db.models.deletion
import image_assets.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='AssetType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(unique=True)),
                ('format', models.CharField(choices=[('jpeg', 'JPEG'), ('png', 'PNG')], max_length=4, verbose_name='Image Format')),
                ('min_width', models.IntegerField(default=0, verbose_name='Min Width')),
                ('min_height', models.IntegerField(default=0, verbose_name='Min Height')),
                ('aspect', models.FloatField(default=0, verbose_name='Aspect')),
                ('accuracy', models.FloatField(default=0.01, verbose_name='Aspect accuracy')),
                ('max_size', models.IntegerField(default=0, verbose_name='Max file size')),
                ('allowed_for', models.ManyToManyField(blank=True, related_name='allowed_asset_types', related_query_name='allowed_asset_types', to='contenttypes.ContentType')),
                ('required_for', models.ManyToManyField(blank=True, related_name='required_asset_types', related_query_name='required_asset_types', to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Asset',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.ImageField(upload_to='', validators=[image_assets.models.AssetType.validate_asset])),
                ('active', models.BooleanField(default=True)),
                ('object_id', models.IntegerField()),
                ('asset_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='image_assets.AssetType')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='DeletedAsset',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.ImageField(upload_to='')),
                ('object_id', models.IntegerField()),
                ('asset_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='image_assets.AssetType')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
                'unique_together': {('content_type', 'object_id')},
            },
        ),
    ]



from django.contrib.contenttypes.fields import GenericRelation
from django.db import models

from image_assets.models import Asset


class Video(models.Model):
    assets = GenericRelation(Asset, blank=True)


class Article(models.Model):
    assets = GenericRelation(Asset, blank=True)



from django import forms
from django.contrib.contenttypes.forms import BaseGenericInlineFormSet
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from image_assets import models


class AssetForm(forms.ModelForm):
    """ Asset form with restricted asset_type choices."""
    asset_type = forms.ModelChoiceField(
        queryset=models.get_asset_type_model().objects.distinct('id'),
        label=_("Asset type")
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        related = self.instance.related
        qs = models.get_asset_type_model().objects.get_for_model(related)
        self.fields["asset_type"].queryset = qs


class AssetFormSet(BaseGenericInlineFormSet):
    model = models.get_asset_model()

    def __init__(self, data=None, files=None, instance=None, save_as_new=False,
                 prefix=None, queryset=None, **kwargs):
        if not kwargs.get('initial') and not (data or files):
            kwargs['initial'] = self.construct_initial(instance)
        super().__init__(data, files, instance, save_as_new, prefix, queryset,
                         **kwargs)

    @staticmethod
    def construct_initial(instance):
        asset_type_model = models.get_asset_type_model()
        initial = []
        for asset_type in asset_type_model.objects.get_required(instance):
            initial.append({'asset_type': asset_type.pk})
        return initial

    def add_fields(self, form, index):
        super().add_fields(form, index)
        qs = models.get_asset_type_model().objects.get_for_model(self.instance)
        form.fields['asset_type'].queryset = qs

    def clean(self):
        super().clean()
        if not self.is_valid():
            return
        duplicated_asset_types = set()
        added_asset_types = set()
        qs = models.get_asset_type_model().objects.get_required(self.instance)
        required_asset_types = set(qs.values_list('pk', flat=True))
        for asset_data in self.cleaned_data:
            active = asset_data.get('active')
            if not active:
                # We only perform extra validation for active assets
                continue
            asset_type = asset_data.get('asset_type')
            if not asset_type:
                # empty field value
                continue
            if asset_type.pk in added_asset_types:
                # there are more than one active assets of this type
                duplicated_asset_types.add(asset_type.pk)
            # track whether there are active assets for all required types
            added_asset_types.add(asset_type.pk)

        errors = []
        missing = required_asset_types - added_asset_types
        if missing:
            qs = models.get_asset_type_model().objects.filter(
                pk__in=missing).values_list('slug', flat=True)
            msg = _("Missing required asset types: %s")
            errors.append(ValidationError(msg % ', '.join(qs)))
        if duplicated_asset_types:
            qs = models.get_asset_type_model().objects.filter(
                pk__in=duplicated_asset_types).values_list('slug', flat=True)
            msg = _("Duplicate active assets for types: %s")
            errors.append(ValidationError(msg % ', '.join(qs)))
        if errors:
            raise ValidationError(errors)



from django.contrib import admin

from image_assets.admin import AssetsInline
from testproject.testapp import models


@admin.register(models.Video)
class VideoAdmin(admin.ModelAdmin):
    inlines = (AssetsInline,)


@admin.register(models.Article)
class ArticleAdmin(admin.ModelAdmin):
    inlines = (AssetsInline,)



import io
from unittest import mock

from PIL import Image
from admin_smoke.tests import AdminTests, AdminBaseTestCase, BaseTestCase
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import SimpleUploadedFile

from image_assets import models as assets_models
from testproject.testapp import admin, models


class VideoBaseTestCase(BaseTestCase):
    """ Common methods for test cases."""

    image: Image.Image

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.image = Image.new('RGB', (60, 30), color='red')
        cls.asset_type = assets_models.AssetType.objects.create(
            slug="video_asset", format=assets_models.AssetType.PNG)
        cls.video_content_type = ContentType.objects.get_for_model(models.Video)
        cls.asset_type.required_for.set([cls.video_content_type])
        cls.video = models.Video.objects.create(pk=23)

    @classmethod
    def create_uploaded_file(cls, image_format='png', filename="asset.png"):
        buffer = io.BytesIO()
        cls.image.save(buffer, format=image_format)
        return SimpleUploadedFile(
            filename, buffer.getvalue(), content_type="image/png")

    def setUp(self):
        super().setUp()
        self.asset = self.video.assets.create(
            asset_type=self.asset_type,
            active=True, image=self.create_uploaded_file())

    # noinspection PyUnusedLocal
    @staticmethod
    def _storage_save_mock(name, content, max_length=None):
        return name


class VideoAssetTypeTestCase(VideoBaseTestCase):
    video_content_type: ContentType

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.allowed_asset_type = assets_models.AssetType.objects.create(
            slug="allowed_asset", format=assets_models.AssetType.PNG)
        cls.allowed_asset_type.allowed_for.set(
            [cls.video_content_type])
        cls.unrelated_asset_type = assets_models.AssetType.objects.create(
            slug="unrelated", format=assets_models.AssetType.PNG)

    def test_required_asset_types(self):
        """ Check fetching required asset types for model or instance."""
        qs = assets_models.AssetType.objects.get_required(self.video)
        # Video has active asset of required asset type
        self.assertEqual(qs.count(), 0)

        qs = assets_models.AssetType.objects.get_required(models.Video)
        # For Video model one required asset type
        self.assertEqual(set(qs), {self.asset_type})

        qs = assets_models.AssetType.objects.get_required(None)
        # Return all asset types
        self.assertEqual(set(qs), {self.asset_type, self.allowed_asset_type,
                                   self.unrelated_asset_type})

        self.update_object(self.asset, active=False)

        qs = assets_models.AssetType.objects.get_required(self.video)
        # Video has inactive asset of required asset type
        self.assertEqual(set(qs), {self.asset_type})

        self.update_object(self.asset, active=True,
                           asset_type=self.allowed_asset_type)

        qs = assets_models.AssetType.objects.get_required(self.video)
        # Video has active asset for another asset type
        self.assertEqual(set(qs), {self.asset_type})



class VideoAdminTestCase(VideoBaseTestCase, AdminTests, AdminBaseTestCase):
    model_admin = admin.VideoAdmin
    model = models.Video
    object_name = 'video'
    prefix = 'image_assets-asset-content_type-object_id'

    def transform_to_new(self, data: dict) -> dict:
        self.reset_inline_data(data, self.prefix, None)
        data[f'{self.prefix}-0-image'] = self.create_uploaded_file()
        return data

    def test_validate_required_asset_type(self):
        """ Without required assets object can't be saved."""
        r = self.client.get(self.change_url)
        data = self.get_form_data_from_response(r)
        url = self.add_url
        data = self.transform_to_new(data)
        for k in list(data):
            if '-0-' in k:
                del data[k]
            elif '-3-' in k:
                data[k.replace('-3-', '-0-')] = data.pop(k)
        data[f'{self.prefix}-TOTAL_FORMS'] = 3

        r = self.client.post(url, data=data)

        self.assertEqual(r.status_code, 200)
        self.assertIsNotNone(self.get_errors_from_response(r))

        self.asset_type.required_for.clear()

        r = self.client.post(url, data=data)

        self.assertEqual(r.status_code, 302)
        self.assertFalse(self.get_errors_from_response(r))

    def test_validate_allowed_asset_type(self):
        """ If asset type is not allowed, object will not be saved."""
        self.asset_type.required_for.clear()
        r = self.client.get(self.change_url)
        data = self.get_form_data_from_response(r)
        url = self.add_url
        data = self.transform_to_new(data)
        data[f'{self.prefix}-0-image'] = self.create_uploaded_file()

        r = self.client.post(url, data=data)

        self.assertEqual(r.status_code, 200)
        self.assertIsNotNone(self.get_errors_from_response(r))

        self.asset_type.allowed_for.set([self.video_content_type])
        data[f'{self.prefix}-0-image'] = self.create_uploaded_file()

        r = self.client.post(url, data=data)

        self.assertFalse(self.get_errors_from_response(r))
        self.assertEqual(r.status_code, 302)

    def test_maintain_single_active_asset(self):
        """ Only one active asset of same asset type is allowed."""
        r = self.client.get(self.change_url)
        data = self.get_form_data_from_response(r)
        data[f'{self.prefix}-1-image'] = self.create_uploaded_file()
        data[f'{self.prefix}-1-asset_type'] = self.asset_type.id

        r = self.client.post(self.change_url, data=data)

        self.assertEqual(r.status_code, 200)
        self.assertTrue(self.get_errors_from_response(r))

        data[f'{self.prefix}-0-active'] = False
        data[f'{self.prefix}-1-image'] = self.create_uploaded_file()

        r = self.client.post(self.change_url, data=data)

        self.assertFalse(self.get_errors_from_response(r))
        self.assertEqual(r.status_code, 302)
        self.assert_object_fields(
            self.asset,
            active=False)
        asset = self.video.assets.last()
        self.assertTrue(asset.active)


class DeletedAssetModelTestCase(VideoBaseTestCase):
    """ Asset deletion handling test case."""

    def setUp(self):
        super().setUp()
        self.delete_patcher = mock.patch(
            'inmemorystorage.InMemoryStorage.delete')
        self.delete_mock = self.delete_patcher.start()

    def tearDown(self):
        super().tearDown()
        self.delete_patcher.stop()

    def test_delete_asset(self):
        """
        Asset file is moved to DeletedAsset instance after asset deletion.
        """
        filename = self.asset.image.name
        self.asset.delete()

        self.delete_mock.assert_not_called()
        deleted = assets_models.DeletedAsset.objects.get()
        self.assertEqual(deleted.image.name, filename)

    def test_purge_deleted_asset(self):
        """
        A file for DeletedAsset is purged on object deletion.
        """
        self.asset.delete()
        deleted = assets_models.DeletedAsset.objects.get()
        filename = deleted.image.name

        deleted.delete()

        self.delete_mock.assert_called_once_with(filename)


class AssetValidationTestCase(VideoBaseTestCase):
    """ Checks image validation for asset type."""

    def assert_validation_not_passed(self):
        try:
            self.asset.full_clean()
        except ValidationError as e:
            invalid_keys = set(e.error_dict)
            expected = {'asset_type', 'content_type', 'object_id'}
            self.assertTrue(invalid_keys - expected)

    def assert_validation_passed(self):
        try:
            self.asset.full_clean()
        except ValidationError as e:
            invalid_keys = set(e.error_dict)
            expected = {'asset_type', 'content_type', 'object_id'}
            self.assertFalse(invalid_keys - expected)

    def test_validate_without_asset_type(self):
        """ Asset object passes image validation if no asset type is set."""
        self.asset.asset_type = None

        self.assert_validation_passed()

        self.asset.asset_type = self.asset_type
        self.asset_type.max_size = 1

        self.assert_validation_not_passed()

    def test_validate_format(self):
        """ Asset image format must correspond asset type format."""
        self.asset_type.format = assets_models.AssetType.JPEG

        self.assert_validation_not_passed()

        self.asset.image = self.create_uploaded_file(
            image_format='jpeg', filename='asset.jpg')

        self.assert_validation_passed()

        # format is checked from image, not filename
        self.asset.image = self.create_uploaded_file(
            image_format='png', filename='asset.jpg')

        self.assert_validation_not_passed()

    def test_validate_min_width(self):
        """ Asset image width must correspond asset type min width."""
        self.asset_type.min_width = self.image.width + 1

        self.assert_validation_not_passed()

        self.asset_type.min_width = self.image.width

        self.assert_validation_passed()

    def test_validate_min_height(self):
        """ Asset image height must correspond asset type min height."""
        self.asset_type.min_height = self.image.height + 1

        self.assert_validation_not_passed()

        self.asset_type.min_height = self.image.height

        self.assert_validation_passed()

    def test_validate_max_size(self):
        """ Asset file size must correspond asset type max size."""
        # zero max size disables file size check
        self.assert_validation_passed()

        self.asset_type.max_size = self.asset.image.size - 1

        self.assert_validation_not_passed()

        self.asset_type.max_size = self.asset.image.size

        self.assert_validation_passed()

    def test_validate_aspect_with_accuracy(self):
        """
        Asset aspect ratio must correspond asset type aspect with accuracy.
        """
        # zero aspect disables aspect check
        self.assert_validation_passed()

        self.asset_type.aspect = 2.1

        self.assert_validation_not_passed()

        self.asset_type.accuracy = 0.1

        self.assert_validation_passed()

        self.asset_type.aspect = 1.8

        self.assert_validation_not_passed()

        self.asset_type.aspect = 1.99
        self.asset_type.accuracy = 0.01

        self.assert_validation_passed()
